# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

from logging import config as logging_config
from . import config


def init_config():
    log_file = config.get("LOG_FILE", "/var/log/funsun.log")
    log_level = config.get("LOG_LEVEL", "INFO")
    log_format = config.get("LOG_FORMAT", "%(asctime)s %(process)d %(levelname)s %(message)s")
    log_rotate = config.get("LOG_ROTATE", "")
    log_backup_Count = config.get("LOG_BACKUP_COUNT", 64)

    if log_file == '-':
        main_handler = {
            'level': log_level,
            'class': 'logging.StreamHandler',
            'formatter': 'main',
        }
    else:
        if log_rotate == "DAY":
            main_handler = {
                'level': log_level,
                'class': 'logging.handlers.TimedRotatingFileHandler',
                'formatter': 'main',
                'filename': log_file,
                'when': "MIDNIGHT",
                'backupCount': log_backup_Count,
            }
        else:
            units = {"B": 1, "K": 1024, "M": 1024 * 1024, "G": 1024 * 1024 * 1024}
            file_bytes = 0

            if log_rotate and log_rotate[-1] in units:
                try:
                    file_bytes = int(float(log_rotate[:-1]) * units[log_rotate[-1]])
                except: pass

            if file_bytes:
                main_handler = {
                    'level': log_level,
                    'class': 'logging.handlers.RotatingFileHandler',
                    'formatter': 'main',
                    'filename': log_file,
                    'maxBytes': file_bytes,
                    'backupCount': log_backup_Count,
                }
            else:
                main_handler = {
                    'level': log_level,
                    'class': 'logging.FileHandler',
                    'formatter': 'main',
                    'filename': log_file,
                }

    log_config={
        "version":1,
        "formatters": {
            "main": {
                "format": log_format,
            }
        },
        "handlers": {
            "logger": main_handler,
        },
        "loggers": {
            "": {
                'handlers': ['logger'],
                'level': log_level
            },
            "tornado.application": {
                'level': log_level
            },
            "tornado.general": {
                'level': log_level
            },
            "tornado.access": {
                'level': log_level
            },
            "tornadis.pool": {
                'handlers': ['logger'],
                'level': "WARNING"
            },
            "tornadis.client": {
                'handlers': ['logger'],
                'level': "WARNING"
            },
            "tornadis.connection": {
                'handlers': ['logger'],
                'level': "WARNING"
            },
        }
    }

    logging_config.dictConfig(log_config)


# -*- coding: utf-8 -*-
# 15/7/4
# create by: snower

import sys

if sys.version_info[0] >= 3:
    is_py3 = True
    string_type = (str, bytes)
    unicode_type = str
    bytes_type = bytes
    number_type = int

    def ensure_bytes(s):
        if isinstance(s, str):
            return s.encode("utf-8")
        return bytes(s)

    def ensure_unicode(s):
        if isinstance(s, bytes):
            return s.decode("utf-8")
        return str(s)
else:
    is_py3 = False
    string_type = (str, unicode)
    unicode_type = unicode
    bytes_type = str
    number_type = (int, long)

    def ensure_bytes(s):
        if isinstance(s, unicode):
            return s.encode("utf-8")
        return str(s)

    def ensure_unicode(s):
        if isinstance(s, unicode):
            return s
        if isinstance(s, str):
            return s.decode("utf-8")
        return str(s).decode("utf-8")

class ExtensionsNotFound(Exception): pass

def load_extensions(name):
    try:
        if "." in name:
            names = name.split(".")
            module = __import__(".".join(names[:-1]), {}, {}, [names[-1]])
            return getattr(module, names[-1])
        else:
            return __import__(name)
    except ImportError:
        return ExtensionsNotFound(name)

def parse_cmd(command, parse_kwargs = True, split_cmd = True):
    cmds = []
    args = []
    carg = []
    is_string = False
    is_kwargs = False
    lc = None

    for c in command:
        if c in ('"', "'"):
            if is_string is False:
                is_string = c
                lc = c
                continue
            elif carg:
                if is_string == c:
                    if lc != '\\':
                        if is_kwargs:
                            args.append((is_kwargs, "".join(carg)))
                            is_kwargs = False
                        else:
                            args.append("".join(carg))
                        carg = []
                        lc = c
                        is_string = False
                        continue
                    else:
                        carg.pop()

        elif c in ('\n', '\t', ' ', '\r'):
            if not carg:
                lc = c
                continue
            elif not is_string:
                if is_kwargs:
                    args.append((is_kwargs, "".join(carg)))
                    is_kwargs = False
                else:
                    args.append("".join(carg))
                carg = []
                lc = c
                continue
        elif c == ';':
            if not is_string:
                if carg:
                    if is_kwargs:
                        args.append((is_kwargs, "".join(carg)))
                        is_kwargs = False
                    else:
                        args.append("".join(carg))

                if args:
                    cmds.append((args[0], tuple(args[1:])))
                args = []
                carg = []
                is_string = False
                lc = None
                continue
        elif c == "=":
            if parse_kwargs and not is_string:
                if carg:
                    is_kwargs = "".join(carg)
                    carg = []
                    lc = c
                    continue

        carg.append(c)
        lc = c

    if carg:
        if is_kwargs:
            args.append((is_kwargs, "".join(carg)))
        else:
            args.append("".join(carg))

    if args:
        if split_cmd:
            cmds.append((args[0], tuple(args[1:])))
        else:
            cmds.append(args)
    return cmds


# -*- coding: utf-8 -*-
# 15/6/10
# create by: snower

import logging
import traceback
from tornado import gen
from .action import Action
from ..error import ActionExecuteRetry
from ..status import forsun_status

__drivers = None

class UnknownActionError(Exception):
    pass

def init_drivers():
    global __drivers
    __drivers = {}

    try:
        from .drivers import shellaction
        __drivers["shell"] = shellaction.ShellAction
        logging.info("action register shell %s", shellaction.ShellAction)
    except Exception as e:
        logging.error("load shell execute error: %s", e)

    try:
        from .drivers import httpaction
        __drivers["http"] = httpaction.HttpAction
        logging.info("action register http %s", httpaction.HttpAction)
    except Exception as e:
        logging.error("load http execute error: %s", e)

    try:
        from .drivers import redisaction
        __drivers["redis"] = redisaction.RedisAction
        logging.info("action register redis %s", redisaction.RedisAction)
    except Exception as e:
        logging.error("load redis execute error: %s", e)

    try:
        from .drivers import thriftaction
        __drivers["thrift"] = thriftaction.ThriftAction
        logging.info("action register thrift %s", thriftaction.ThriftAction)
    except Exception as e:
        logging.error("load thrift execute error: %s", e)

    try:
        from .drivers import beanstalkaction
        __drivers["beanstalk"] = beanstalkaction.BeanstalkAction
        logging.info("action register beanstalk %s", beanstalkaction.BeanstalkAction)
    except Exception as e:
        logging.error("load beanstalk execute error: %s", e)

    try:
        from .drivers import mysqlaction
        __drivers["mysql"] = mysqlaction.MysqlAction
        logging.info("action register mysql %s", mysqlaction.MysqlAction)
    except Exception as e:
        logging.error("load mysql execute error: %s", e)

def register_action(name, cls):
    if issubclass(cls, Action):
        __drivers[name] = cls
        logging.info("action register %s %s", name, cls)
        return True
    return False

def get_driver(action):
    if __drivers is None:
        init_drivers()
    try:
        driver = __drivers[action]
    except:
        raise UnknownActionError()
    return driver

def get_driver_names():
    if __drivers is None:
        return []
    return __drivers.keys()

@gen.coroutine
def execute(ts, plan):
    driver_cls = get_driver(plan.action)
    driver = driver_cls(ts, plan, plan.action, plan.params)
    try:
        yield driver.execute()
        succed = True
    except ActionExecuteRetry as e:
        raise e
    except Exception as e:
        forsun_status.action_executed_error_count += 1
        logging.error("action %s %s %s %s execute error: %s\n%s", plan.key, driver.action, driver.ts, driver.params, e, traceback.format_exc())
        succed = False
    raise gen.Return(succed)


# -*- coding: utf-8 -*-
# 15/6/8
# create by: snower

import logging
from .. import config
from .store import Store

class UnknownStoreDriverError(Exception):
    pass

__stores = {}

def init_stores():
    try:
        from .drivers.mem import MemStore
        __stores['mem'] = MemStore
        logging.info("store register mem %s", MemStore)
    except Exception as e:
        logging.error("store load mem error: %s", e)

    try:
        from .drivers.redis import RedisStore
        __stores['redis'] = RedisStore
        logging.info("store register redis %s", RedisStore)
    except Exception as e:
        logging.error("store load redis error: %s", e)

def register_store(name, cls):
    if issubclass(cls, Store):
        __stores[name] = cls
        logging.info("store register %s %s", name, cls)
        return True
    return False

def get_store(*args, **kwargs):
    driver = config.get("STORE_DRIVER", "mem")
    if not __stores:
        init_stores()

    if driver not in __stores:
        raise UnknownStoreDriverError()
    return __stores[driver](*args, **kwargs)

def get_store_names():
    return __stores.keys()


# -*- coding: utf-8 -*-
# 18/3/14
# create by: snower

import time
import json
from tornado import gen
from tornado.web import RequestHandler as BaseRequestHandler
from tornado.web import HTTPError
from tornado.httpserver import HTTPServer as BaseHTTPServer
from tornado.web import Application as BaseApplication
from ..plan import Plan
from ..utils import parse_cmd, unicode_type
from ..status import forsun_status
from ..error import ForsunPlanError, RequiredArgumentError, NotFoundPlanError

def execute(func):
    @gen.coroutine
    def _(self, *args, **kwargs):
        forsun_status.http_requesting_count += 1
        try:
            result = yield func(self, *args, **kwargs)
            data = {
                "errcode": 0,
                "errmsg": "",
                "data": result
            }
        except ForsunPlanError as e:
            data = {
                "errcode": e.code,
                "errmsg": e.message,
                "data": {}
            }
        except Exception as e:
            forsun_status.http_requested_error_count += 1
            data = {
                "errcode": 1001,
                "errmsg": u"未知错误",
                "data": {}
            }
        finally:
            forsun_status.http_requested_count += 1
            forsun_status.http_requesting_count -= 1
        self.finish(data)
    return _

class RequestHandler(BaseRequestHandler):
    is_timeout = False

    def prepare(self):
        if self.request.method.lower() in ("post", "put"):
            content_type = self.request.headers.get("Content-Type", "application/json")
            if content_type.startswith("application/json"):
                try:
                    if not isinstance(self.request.body, unicode_type):
                        body = unicode_type(self.request.body, "utf-8")
                    else:
                        body = self.request.body
                    self.request.body_arguments = json.loads(body)
                    if isinstance(self.request.body_arguments, dict):
                        if self.request.method.lower() == "put":
                            self.request.body_arguments["method"] = "createTimeout"
                        else:
                            self.request.body_arguments["method"] = "create"
                        self.request.body_arguments = [self.request.body_arguments]
                except:
                    raise HTTPError(400, u"请求数据类型application/json解析失败")
            elif content_type.startswith("application/crontab"):
                try:
                    if not isinstance(self.request.body, unicode_type):
                        body = unicode_type(self.request.body, "utf-8")
                    else:
                        body = self.request.body
                    self.request.body_arguments = self.parse_cmd(body)
                except:
                    import traceback
                    traceback.print_exc()
                    raise HTTPError(400, u"请求数据类型接受application/crontab解析失败")
            else:
                raise HTTPError(400, u"未知数据类型")


    def plan_to_dict(self, plan):
        return {
            "key": plan.key,
            "second": plan.second,
            "minute": plan.minute,
            "hour": plan.hour,
            "day": plan.day,
            "month": plan.month,
            "week": plan.week,
            "status": plan.status,
            "count": plan.count,
            "is_time_out": plan.is_time_out,
            "next_time": plan.next_time,
            "current_count": plan.current_count,
            "last_timeout": plan.last_timeout,
            "created_time": plan.created_time,
            "action": plan.action,
            "params": plan.params,
        }

    def parse_cmd(self, body):
        arguments = []
        for b in body.split("\n"):
            cmds = parse_cmd(b, True, False)
            for cmd in cmds:
                key, seconds, minutes, hours, days, months, weeks, action, params_str = cmd

                is_timeout = False
                for ct in [seconds, minutes, hours, days, months, weeks]:
                    if ct.startswith("*/"):
                        is_timeout = True
                        break

                params = {}
                for cmd, args in parse_cmd(params_str):
                    if isinstance(cmd, tuple):
                        params[cmd[0]] = cmd[1]
                    else:
                        params[cmd] = args

                if is_timeout:
                    def parse_time(ct, count):
                        if not ct.startswith("*/"):
                            return 0, count

                        ct = ct.split("/")
                        return (int(ct[1]), int(ct[2])) if len(ct) >= 3 else (int(ct[1]), count)

                    count = 1
                    seconds, count = parse_time(seconds, count)
                    minutes, count = parse_time(minutes, count)
                    hours, count = parse_time(hours, count)
                    days, count = parse_time(days, count)
                    months, count = parse_time(months, count)
                    weeks, count = parse_time(weeks, count)
                else:
                    count = 0
                    seconds = int(seconds) if seconds != "*" else -1
                    minutes = int(minutes) if minutes != "*" else -1
                    hours = int(hours) if hours != "*" else -1
                    days = int(days) if days != "*" else -1
                    months = int(months) if months != "*" else -1
                    weeks = int(weeks) if weeks != "*" else -1

                arguments.append({
                    "method": "createTimeout" if is_timeout else "create",
                    "key": key,
                    "seconds": seconds,
                    "minute": minutes,
                    "hour": hours,
                    "day": days,
                    "month": months,
                    "week": weeks,
                    "count": count,
                    "action": action,
                    "params": params,
                })
        return arguments

    @gen.coroutine
    def create_plan(self):
        plans = []
        for arguments in self.request.body_arguments:
            key = arguments.get("key", "")
            seconds = int(arguments.get("seconds", -1))
            minute = int(arguments.get("minute", -1))
            hour = int(arguments.get("hour", -1))
            day = int(arguments.get("day", -1))
            month = int(arguments.get("month", -1))
            week = int(arguments.get("week", -1))
            count = int(arguments.get("count", 0))
            action = arguments.get("action", "shell")
            params = arguments.get("params", {})
            if not isinstance(params, dict):
                continue

            if arguments.get("method", "createTimeout") == "createTimeout":
                plan = Plan(key, seconds, minute, hour, day, month, week, is_time_out=True, count=count, action=action, params=params,
                            created_time=time.mktime(time.gmtime()))
                yield self.application.forsun.create_plan(plan)
            else:
                plan = Plan(key, seconds, minute, hour, day, month, week, is_time_out=False, action=action, params=params,
                        created_time=time.mktime(time.gmtime()))
                yield self.application.forsun.create_plan(plan)
            plans.append(self.plan_to_dict(plan))
        if not plans:
            raise RequiredArgumentError()
        raise gen.Return(plans[0] if len(plans) == 1 else plans)

class PingRequestHandler(RequestHandler):
    @execute
    @gen.coroutine
    def get(self):
        raise gen.Return({})

class PlanRequestHandler(RequestHandler):
    @execute
    @gen.coroutine
    def get(self):
        key = self.get_query_argument("key")
        if not key:
            raise RequiredArgumentError("key")

        plan = yield self.application.forsun.get_pan(key)
        if not plan:
            raise NotFoundPlanError()
        raise gen.Return(self.plan_to_dict(plan))

    @execute
    @gen.coroutine
    def post(self):
        plan = yield self.create_plan()
        raise gen.Return(plan)

    @execute
    @gen.coroutine
    def put(self):
        plan = yield self.create_plan()
        raise gen.Return(plan)

    @execute
    @gen.coroutine
    def delete(self):
        key = self.get_query_argument("key")
        if not key:
            raise RequiredArgumentError("key")

        plan = yield self.application.forsun.remove_plan(key)
        if not plan:
            raise NotFoundPlanError()
        raise gen.Return(self.plan_to_dict(plan))

class TimeRequestHandler(RequestHandler):
    @execute
    @gen.coroutine
    def get(self):
        timestamp = self.get_query_argument("timestamp", default=0)

        current_timestamp = yield self.application.forsun.get_current_time()
        if not timestamp:
            timestamp = current_timestamp
        else:
            try:
                timestamp = int(timestamp)
            except:
                raise RequiredArgumentError("timestamp")

        plans = yield self.application.forsun.get_time_plans(timestamp)
        raise gen.Return({
            "plans": [self.plan_to_dict(plan) for plan in plans],
            "current_timestamp": current_timestamp,
        })

class KeysRequestHandler(RequestHandler):
    @execute
    @gen.coroutine
    def get(self):
        prefix = self.get_query_argument("prefix", default="")

        keys = yield self.application.forsun.get_keys(prefix)
        raise gen.Return(keys)

class InfoRequestHandler(RequestHandler):
    @execute
    @gen.coroutine
    def get(self):
        info =  forsun_status.get_info()
        raise gen.Return(info)

class HTTPServer(BaseHTTPServer):
    def handle_stream(self, *args, **kwargs):
        r = super(HTTPServer, self).handle_stream(*args, **kwargs)
        forsun_status.http_connecting_count+=1
        forsun_status.http_connected_count += 1
        return r

    def on_close(self, *args, **kwargs):
        r = super(HTTPServer, self).on_close(*args, **kwargs)
        forsun_status.http_connecting_count -= 1
        return r

class Application(BaseApplication):
    def __init__(self, forsun, *args, **kwargs):
        self.forsun = forsun

        urls = (
            (r'^/v1/ping$', PingRequestHandler),
            (r'^/v1/plan$', PlanRequestHandler),
            (r'^/v1/time$', TimeRequestHandler),
            (r'^/v1/keys$', KeysRequestHandler),
            (r'^/v1/info$', InfoRequestHandler),
        )

        super(Application, self).__init__(urls, *args, **kwargs)


# -*- coding: utf-8 -*-
# 18/1/25
# create by: snower

import os
import time
import logging
import math
from collections import defaultdict
from tornado.ioloop import IOLoop
import msgpack
from tornado import gen
from ...plan import Plan
from ..store import Store
from ...utils import unicode_type
from ... import config

class MemStore(Store):
    def __init__(self, *args, **kwargs):
        super(MemStore, self).__init__(*args, **kwargs)

        self.ioloop = IOLoop.current()
        self.plans = {}
        self.time_plans = defaultdict(dict)
        self.current_time = int(time.mktime(time.gmtime()))
        self.store_file = config.get("STORE_MEM_STORE_FILE", "")
        if not self.store_file:
            self.store_file = os.path.join(os.path.expanduser('~'), ".forsun.dump")
        self.store_file = os.path.abspath(self.store_file)
        self.store_time_rate = config.get("STORE_MEM_TIME_RATE", 1)
        self.store_waited = False

        self.load()
        logging.info("use mem store %s", self.store_file)

    def load(self):
        try:
            with open(self.store_file, "rb") as fp:
                data = fp.read()
                data = msgpack.loads(data)
                self.plans = {}
                for plan in data.get(b"plans", []):
                    plan = Plan.loads(plan)
                    self.plans[plan.key] = plan

                self.time_plans = defaultdict(dict)
                for key, plans in data.get(b"time_plans", {}).items():
                    self.time_plans[key] = {}
                    for plan_key in (plans or []):
                        if not isinstance(plan_key, unicode_type):
                            plan_key = unicode_type(plan_key, "utf-8")

                        if plan_key in self.plans:
                            self.time_plans[key][plan_key] = self.plans[plan_key]

                current_time = data.get(b"current_time")
                self.current_time = current_time if current_time and current_time > 0 else self.current_time
            logging.info("store mem load session %s", self.store_file)
        except Exception as e:
            logging.error("store mem load session error: %s %s", self.store_file, e)

    def store(self, saved_current_time = False):
        try:
            with open(self.store_file, "wb") as fp:
                data = msgpack.dumps({
                    b"plans": [plan.dumps() for key, plan in self.plans.items()],
                    b"time_plans": {ts: list(plans.keys()) for ts, plans in self.time_plans.items()},
                    b"current_time": self.current_time if saved_current_time else 0,
                })
                fp.write(data)

            if saved_current_time:
                logging.info("store mem save session %s", self.store_file)
        except Exception as e:
            logging.error("store mem save session error: %s %s", self.store_file, e)

    @gen.coroutine
    def uninit(self):
        self.store(True)

    @gen.coroutine
    def set_current(self, current_time):
        self.current_time = current_time
        raise gen.Return(self.current_time)

    @gen.coroutine
    def get_current(self):
        raise gen.Return(self.current_time)

    @gen.coroutine
    def set_plan(self, plan):
        self.plans[plan.key] = plan

        if not self.store_waited:
            timeout = math.sqrt(len(self.plans)) * self.store_time_rate
            def on_store():
                self.store()
                self.store_waited = False
            self.ioloop.call_later(timeout, on_store)
            self.store_waited = True
        raise gen.Return(plan)

    @gen.coroutine
    def remove_plan(self, key):
        if key in self.plans:
            plan = self.plans.pop(key)

            if not self.store_waited:
                timeout = math.sqrt(len(self.plans)) * self.store_time_rate

                def on_store():
                    self.store()
                    self.store_waited = False

                self.ioloop.call_later(timeout, on_store)
                self.store_waited = True

            raise gen.Return(plan)
        raise gen.Return(None)

    @gen.coroutine
    def get_plan(self, key):
        if key in self.plans:
            raise gen.Return(self.plans[key])
        raise gen.Return(None)

    @gen.coroutine
    def add_time_plan(self, next_time, key):
        self.time_plans[next_time][key] = 0
        raise gen.Return(True)

    @gen.coroutine
    def get_time_plan(self, next_time, key):
        status = self.time_plans[next_time].get(key, 0)
        raise gen.Return(status)

    @gen.coroutine
    def set_time_plan(self, next_time, key, status):
        last_status = self.time_plans[next_time].get(key, 0)
        self.time_plans[next_time][key] = status
        raise gen.Return(last_status)

    @gen.coroutine
    def remove_time_plan(self, next_time, key):
        if key in self.time_plans[next_time]:
            raise gen.Return(self.time_plans[next_time].pop(key))
        raise gen.Return(None)

    @gen.coroutine
    def get_time_plans(self, ts):
        raise gen.Return(self.time_plans[ts].keys())

    @gen.coroutine
    def delete_time_plans(self, ts):
        if ts in self.time_plans:
            raise gen.Return(self.time_plans.pop(ts))
        raise gen.Return(None)

    @gen.coroutine
    def get_plan_keys(self, prefix = ""):
        prefix = prefix.rstrip("\n\t *")
        keys = []
        for key in self.plans:
            if key.startswith(prefix):
                keys.append(key)
        raise gen.Return(keys)


# -*- coding: utf-8 -*-
# 18/2/7
# create by: snower

import time
import logging
from tornado import gen
from tormysql.helpers import ConnectionPool
from ..action import Action, ExecuteActionError
from ... import config
from ...error import ActionExecuteRetry

class MysqlAction(Action):
    client_pools = {}

    def __init__(self, *args, **kwargs):
        super(MysqlAction, self).__init__(*args, **kwargs)

        self.start_time  = time.time()

    def get_client(self, host, port, db, user, passwd, max_connections):
        key = "%s:%s:%s:%s" % (host, port, db, user)
        if key not in self.client_pools:
            wait_connection_timeout = int(config.get("ACTION_MYSQL_WAIT_CONNECTION_TIMEOUT", 7200))
            idle_seconds = int(config.get("ACTION_MYSQL_IDLE_SECONDS", 120))
            self.__class__.client_pools[key] = ConnectionPool(host = host, port = port, db = db,
                                                              user = user, passwd = passwd,
                                                              max_connections = max_connections,
                                                              wait_connection_timeout = wait_connection_timeout,
                                                              idle_seconds = idle_seconds,
                                                            )
        return self.__class__.client_pools[key]

    @gen.coroutine
    def execute(self, *args, **kwargs):
        if len(self.params) < 4:
            raise ExecuteActionError("redis params is empty")

        host = self.params.get("host", "127.0.0.1")
        port = int(self.params.get("port", 3306))
        db = self.params.get("db", 'mysql')
        user = self.params.get("user", config.get("ACTION_MYSQL_USER", 'root'))
        passwd = self.params.get("passwd", config.get("ACTION_MYSQL_PASSWD", ''))
        max_connections = int(self.params.get("max_connections", config.get("ACTION_MYSQL_MAX_CONNECTIONS", 8)))
        sql = self.params.get("sql")

        if not sql:
            logging.error("mysql action execute error %s sql is empty", self.plan.key)
            raise gen.Return(None)

        try:
            client = self.get_client(host, port, db, user, passwd, max_connections)
            tx = yield client.begin()
            try:
                yield tx.execute(sql)
            except Exception as e:
                yield tx.rollback()
                logging.error("mysql action execute error '%s' %s %s:%s/%s '%s' '%s' %.2fms", self.plan.key, user, host,
                              port, db, sql, e, (time.time() - self.start_time) * 1000)
            else:
                yield tx.commit()
                logging.debug("mysql action execute '%s' %s %s:%s/%s '%s' %.2fms", self.plan.key, user, host, port, db,
                              sql, (time.time() - self.start_time) * 1000)
        except Exception as e:
            logging.error("mysql action execute error '%s' %s %s:%s/%s '%s' '%s' %.2fms", self.plan.key, user, host,
                          port, db, sql, e, (time.time() - self.start_time) * 1000)
            raise ActionExecuteRetry()


# -*- coding: utf-8 -*-
# 18/8/29
# create by: snower

import sys
import os
import datetime
import tzlocal
import psutil
from .version import version

current_tz = tzlocal.get_localzone()

class ForsunStatus(object):
    def __init__(self):
        self.start_time = datetime.datetime.now(current_tz)
        self.connecting_count = 0
        self.connected_count = 0
        self.requesting_count = 0
        self.requested_count = 0
        self.requested_error_count = 0
        self.http_connecting_count = 0
        self.http_connected_count = 0
        self.http_requesting_count = 0
        self.http_requested_count = 0
        self.http_requested_error_count = 0
        self.created_count = 0
        self.create_timeouted_count = 0
        self.removed_count = 0
        self.plan_count = 0
        self.timeout_handling_count = 0
        self.timeout_handled_count = 0
        self.action_executing_count = 0
        self.action_executed_count = 0
        self.action_executed_error_count = 0
        self.action_retried_count = 0
        self.timer_loop_count = 0
        self.timer_loop_ready_error_count = 0
        self.customs = {}

    def inc(self, key, value = 1):
        if key not in self.customs:
            self.customs[key] = value
        else:
            self.customs[key] += value
        
    def get_info(self):
        from .store import get_store_names
        from .action import get_driver_names
        from .timer import current
        from . import config

        process = psutil.Process(os.getpid())
        cpu_times = process.cpu_times()
        memory_info = process.memory_info()

        info = {
            "python_version": sys.version.replace("\n", " "),
            "forsun_version": version,
            "start_time": str(self.start_time),
            "cpu_user": str(cpu_times.user),
            "cpu_system": str(cpu_times.system),
            "mem_rss": str(memory_info.rss),
            "mem_vms": str(memory_info.vms),
            "current_time": str(current()),

            "stores": ";".join(get_store_names()),
            "current_store": config.get("STORE_DRIVER", ""),
            "actions": ";".join(get_driver_names()),
            "bind_port": "%s:%s" % (config.get("BIND_ADDRESS", ""), config.get("PORT", "")),
            "http_bind_port": config.get("HTTP_BIND", ""),
            "extensions": ";".join(config.get("EXTENSIONS", [])),

            "connecting_count": str(self.connecting_count),
            "connected_count": str(self.connected_count),
            "requesting_count": str(self.requesting_count),
            "requested_count": str(self.requested_count),
            "requested_error_count": str(self.requested_error_count),
            "http_connecting_count": str(self.http_connecting_count),
            "http_connected_count": str(self.http_connected_count),
            "http_requesting_count": str(self.http_requesting_count),
            "http_requested_count": str(self.http_requested_count),
            "http_requested_error_count": str(self.http_requested_error_count),
            "created_count": str(self.created_count),
            "create_timeouted_count": str(self.create_timeouted_count),
            "removed_count": str(self.removed_count),
            "plan_count": str(self.plan_count),
            "timeout_handling_count": str(self.timeout_handling_count),
            "timeout_handled_count": str(self.timeout_handled_count),
            "action_executing_count": str(self.action_executing_count),
            "action_executed_count": str(self.action_executed_count),
            "action_executed_error_count": str(self.action_executed_error_count),
            "action_retried_count": str(self.action_retried_count),

            "timer_loop_count": str(self.timer_loop_count),
            "timer_loop_ready_error_count": str(self.timer_loop_ready_error_count),
        }

        for key, value in self.customs.items():
            info[key] = str(value)

        return info

forsun_status = ForsunStatus()


# -*- coding: utf-8 -*-
# 15/7/2
# create by: snower

from tornado import gen
from thrift.transport.TSocket import TSocket
from thrift.transport.TTransport import TBufferedTransport
from thrift.protocol.TBinaryProtocol import TBinaryProtocol, TBinaryProtocolFactory
from torthrift.pool import TStreamPool
from torthrift.client import PoolClient
from .client.Forsun import Client
from .client.Forsun import ForsunPlanError

class ThriftClient(object):
    def __init__(self, port=6458, host="127.0.0.1"):
        self.host = host
        self.port = port
        self.client = None
        self.transport = None

    def connect(self):
        self.transport = TSocket(self.host, self.port)
        self.transport = TBufferedTransport(self.transport)
        protocol = TBinaryProtocol(self.transport)
        self.client = Client(protocol)
        self.transport.open()

    def execute(self, name, *args, **kwargs):
        if self.client is None:
            self.connect()
        result = getattr(self.client, name)(*args, **kwargs)
        return result

    def create(self, key, second, minute = -1, hour = -1, day = -1, month = -1, week = -1, action="shell", params={}):
        return self.execute("create", key, second, minute, hour, day, month, week, action, params)

    def create_timeout(self, key, second, minute = -1, hour = -1, day = -1, month = -1, week = -1, count=1, action="shell", params={}):
        return self.execute("createTimeout", key, second, minute, hour, day, month, week, count, action, params)

    def remove(self, key):
        return self.execute("remove", key)

    def get(self, key):
        return self.execute("get", key)

    def get_current(self):
        return self.execute("getCurrent")

    def get_time(self, timestamp):
        return self.execute("getTime", timestamp)

    def get_keys(self, prefix=''):
        return self.execute("getKeys", prefix)

    def info(self):
        return self.execute("info")

    def __del__(self):
        if self.client:
            self.transport.close()
            self.transport = None
            self.client = None

class TorThriftClient(object):
    def __init__(self, port=6458, host="127.0.0.1", max_stream=4):
        self.host = host
        self.port = port
        self.max_stream = max_stream

        self.pool = self.init_pool()

    def init_pool(self):
        return PoolClient(Client, TStreamPool(self.host, self.port, max_stream=self.max_stream), TBinaryProtocolFactory())

    @gen.coroutine
    def create(self, key, second, minute = -1, hour = -1, day = -1, month = -1, week = -1, action="shell", params={}):
        res = yield self.pool.create(key, second, minute, hour, day, month, week, action, params)
        raise gen.Return(res)

    @gen.coroutine
    def create_timeout(self, key, second, minute = -1, hour = -1, day = -1, month = -1, week = -1, count=1, action="shell", params={}):
        res = yield self.pool.createTimeout(key, second, minute, hour, day, month, week, count, action, params)
        raise gen.Return(res)

    @gen.coroutine
    def remove(self, key):
        res = yield self.pool.remove(key)
        raise gen.Return(res)

    @gen.coroutine
    def get(self, key):
        res = yield self.pool.get(key)
        raise gen.Return(res)

    @gen.coroutine
    def get_current(self):
        res = yield self.pool.getCurrent()
        raise gen.Return(res)

    @gen.coroutine
    def get_time(self, timestamp):
        res = yield self.pool.getTime(timestamp)
        raise gen.Return(res)

    @gen.coroutine
    def get_keys(self, prefix=''):
        res = yield self.pool.getKeys(prefix)
        raise gen.Return(res)

    @gen.coroutine
    def info(self):
        res = yield self.pool.info()
        raise gen.Return(res)


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

from __future__ import absolute_import, division, print_function, with_statement

import time
import datetime
import argparse
import pytz
import tzlocal
from thrift.transport.TTransport import TTransportException
from ..version import version
from ..clients import ThriftClient, ForsunPlanError
from ..utils import parse_cmd, string_type, unicode_type

parser = argparse.ArgumentParser(description='High-performance timing scheduling service')
parser.add_argument('--host', dest='host', default="127.0.0.1", help='host (default: 127.0.0.1)')
parser.add_argument('--port', dest='port', default=6458, type=int, help='port (default: 6458)')
parser.add_argument('--exe', dest='execute', default='', type=str, help='execute cmd (default: )')
parser.add_argument('cmd', default='', type=str, nargs=argparse.OPTIONAL, help='execute cmd (default: )')

client = None

def print_plan(plan):
    times = []
    if plan.is_time_out:
        times.append("*" if plan.second == 0 else "*/%s/%s" % (plan.second, plan.count))
        times.append("*" if plan.minute == 0 else "*/%s/%s" % (plan.minute, plan.count))
        times.append("*" if plan.hour == 0 else "*/%s/%s" % (plan.hour, plan.count))
        times.append("*" if plan.day == 0 else "*/%s/%s" % (plan.day, plan.count))
        times.append("*" if plan.month == 0 else "*/%s/%s" % (plan.month, plan.count))
        times.append("*" if plan.week == -1 else "*/%s/%s" % (plan.week, plan.count))
    else:
        times.append("*" if plan.second == -1 else str(plan.second))
        times.append("*" if plan.minute == -1 else str(plan.minute))
        times.append("*" if plan.hour == -1 else str(plan.hour))
        times.append("*" if plan.day == -1 else str(plan.day))
        times.append("*" if plan.month == -1 else str(plan.month))
        times.append("*" if plan.week == -1 else str(plan.week))

    params = ";".join(["%s=%s" % (key, ("'%s'" % value) if isinstance(value, string_type) else value) for key, value in plan.params.items()])

    print(str(datetime.datetime.fromtimestamp(plan.next_time).replace(tzinfo=pytz.UTC).astimezone(tzlocal.get_localzone())),
          plan.key, " ".join(times), plan.action, '"' + params + '"')

def cmd_help(*args):
    print("help - show help doc")
    print("exit - exit the cmd")

    print("ls - show prfix key plans (ls [prefix or *])")
    print("current - show current time plans (current)")
    print("time - show time plans (time [timestamp or datetime])")
    print("get - show plan info (get [key])")
    print("rm - remove plan (rm [key])")
    print("set - add crontab plan (set [key] [seconds] [minutes] [hours] [days] [months] [weeks] [cmd] [args])")

def cmd_exit(*args):
    exit(0)

def cmd_version(*args):
    print(version)

def cmd_ls(prefix = None, *args):
    keys = client.get_keys(prefix or '')
    for key in keys:
        try:
            plan = client.get(key)
            print_plan(plan)
        except ForsunPlanError:
            pass

def cmd_current(*args):
    plans = client.get_current()
    for plan in plans:
        print_plan(plan)

def cmd_time(ts = None, *args):
    if not ts:
        ts = int(time.mktime(time.gmtime()))
    elif ts.isdigit():
        ts = int(ts)
    else:
        ts = int(time.mktime(datetime.datetime.strptime(ts, "%Y-%m-%d %H:%M:%S").replace(tzinfo=pytz.UTC).timetuple()))
    plans = client.get_time(ts)
    for plan in plans:
        print_plan(plan)

def cmd_get(key, *args):
    plan = client.get(key)
    print_plan(plan)

def cmd_rm(key, *args):
    client.remove(key)
    print('success')

def cmd_set(key, seconds, minutes, hours, days, months, weeks, action, params_str, *args):
    is_timeout = False
    for ct in [seconds, minutes, hours, days, months, weeks]:
        if ct.startswith("*/"):
            is_timeout = True
            break

    if action in ("sh", "bash"):
        action = "shell"

    params = {}
    for cmd, args in parse_cmd(params_str):
        if isinstance(cmd, tuple):
            params[cmd[0]] = cmd[1]
        else:
            params[cmd] = args

    if action == "shell":
        if "cmd" not in params:
            params = {"cmd": params_str}
    elif action == "http":
        if "url" not in params:
            params = {"url": params_str}

    if is_timeout:
        def parse_time(ct, count):
            if not ct.startswith("*/"):
                return 0, count

            ct = ct.split("/")
            return (int(ct[1]), int(ct[2])) if len(ct) >= 3 else (int(ct[1]), count)

        count = 1
        seconds, count = parse_time(seconds, count)
        minutes, count = parse_time(minutes, count)
        hours, count = parse_time(hours, count)
        days, count = parse_time(days, count)
        months, count = parse_time(months, count)
        weeks, count = parse_time(weeks, count)

        plan = client.create_timeout(key, seconds, minutes, hours, days, months, weeks or -1, count, action, params)
    else:
        seconds = int(seconds) if seconds != "*" else -1
        minutes = int(minutes) if minutes != "*" else -1
        hours = int(hours) if hours != "*" else -1
        days = int(days) if days != "*" else -1
        months = int(months) if months != "*" else -1
        weeks = int(weeks) if weeks != "*" else -1

        plan = client.create(key, seconds, minutes, hours, days, months, weeks, action, params)

    print_plan(plan)

def cmd_info(*args):
    info = client.info()

    keys = [
        "python_version",
        "forsun_version",
        "start_time",
        "cpu_user",
        "cpu_system",
        "mem_rss",
        "mem_vms",
        "current_time",
        "",

        "stores",
        "current_store",
        "actions",
        "bind_port",
        "http_bind_port",
        "extensions",
        "",

        "connecting_count",
        "connected_count",
        "requesting_count",
        "requested_count",
        "requested_error_count",
        "",

        "http_connecting_count",
        "http_connected_count",
        "http_requesting_count",
        "http_requested_count",
        "http_requested_error_count",
        "",

        "created_count",
        "create_timeouted_count",
        "removed_count",
        "plan_count",
        "timeout_handling_count",
        "timeout_handled_count",
        "",

        "action_executing_count",
        "action_executed_count",
        "action_executed_error_count",
        "action_retried_count",
        "",

        "timer_loop_count",
        "timer_loop_ready_error_count",
        "",
    ]

    for key in keys:
        if key in ("mem_rss", "mem_vms"):
            value, unit = int(info.get(key), 0), "B"
            for unit in ["B", "K", "M", "G"]:
                if value < 1024:
                    break
                value = float(value) / 1024.0

            print(key + ":\t" + ("%.2f" % value) + unit)
            continue

        if key == "current_time":
            print(key + ":\t" + str(datetime.datetime.fromtimestamp(int(info.get(key), 0)).replace(tzinfo=pytz.UTC).astimezone(tzlocal.get_localzone())))
            continue

        print((key + ":\t" + info.get(key)) if key else "")

def main():
    global client

    args = parser.parse_args()
    client = ThriftClient(args.port, args.host)

    try:
        client.connect()
    except TTransportException as e:
        print('connect forsun server fail: ', e.message)
        exit(1)

    CMDS = {
        'help': cmd_help,
        'exit': cmd_exit,
        'version': cmd_version,

        'ls': cmd_ls,
        'current': cmd_current,
        'time': cmd_time,
        'get': cmd_get,
        'rm': cmd_rm,
        'set': cmd_set,
        'info': cmd_info,
    }

    if args.execute or args.cmd:
        cmds = parse_cmd(args.cmd or args.execute)
        for cmd, params in cmds:
            if cmd not in CMDS:
                print('Unknown cmd ', cmd)
                exit(1)
            else:
                try:
                    CMDS[cmd](*params)
                except KeyboardInterrupt:
                    exit(0)
                except ForsunPlanError as e:
                    print('cmd ForsunPlanError: ', e.code, e.message)
                    exit(1)
                except Exception as e:
                    print('cmd error: ', e)
                    exit(1)
        exit(0)

    print('Welcome to the Forsun', version)
    print("Type 'help' for help")
    print('')

    while True:
        try:
            command = input("forsun> ")
            cmds = parse_cmd(command)
            for cmd, params in cmds:
                if cmd not in CMDS:
                    print('Unknown cmd ', cmd)
                else:
                    try:
                        CMDS[cmd](*params)
                        print('')
                    except KeyboardInterrupt:
                        break
                    except ForsunPlanError as e:
                        print('cmd ForsunPlanError: ', e.code, e.message)
                    except Exception as e:
                        print('cmd error: ', e)
        except KeyboardInterrupt:
            break

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
# 15/6/10
# create by: snower

import logging
import threading
from tornado.ioloop import IOLoop, asyncio
from thrift.protocol.TBinaryProtocol import TBinaryProtocolAcceleratedFactory
from torthrift.transport import TIOStreamTransportFactory
from torthrift.server import TTornadoServer as BaseTTornadoServer
from .processor.Forsun import Processor
from .handler import Handler
from .http import HTTPServer, Application
from .. import timer
from ..status import forsun_status
from .. import config

class TTornadoServer(BaseTTornadoServer):
    def process(self, *args, **kwargs):
        try:
            forsun_status.connecting_count += 1
            forsun_status.connected_count += 1
            return super(TTornadoServer, self).process(*args, **kwargs)
        finally:
            forsun_status.connecting_count -= 1

class Server(object):
    def __init__(self, forsun):
        self.forsun = forsun
        self.server = None
        self.http_server = None
        self.thread = None

    def serve_thrift(self):
        handler = Handler(self.forsun)
        processor = Processor(handler)
        tfactory = TIOStreamTransportFactory()
        protocol = TBinaryProtocolAcceleratedFactory()

        bind_address = config.get("BIND_ADDRESS", "127.0.0.1")
        port = config.get("PORT", 6458)
        self.server = TTornadoServer(processor, tfactory, protocol)
        self.server.bind(port, bind_address)
        self.server.start(1)
        logging.info("starting server by %s:%s", bind_address, port)

    def serve_http(self):
        http_bind = config.get("HTTP_BIND")
        if not http_bind:
            return

        (address, port) = http_bind.split(":")
        application = Application(self.forsun, debug=False, autoreload=False)
        self.http_server = HTTPServer(application, xheaders=True)
        self.http_server.bind(int(port), address)
        self.http_server.start(1)
        logging.info("starting http server by %s", http_bind)

    def start(self, init_callback):
        def _():
            try:
                if asyncio is not None:
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)

                self.serve_thrift()
                self.serve_http()
                ioloop = IOLoop.instance()
                ioloop.add_callback(init_callback)
                ioloop.start()
            except Exception as e:
                logging.error("server error: %s", e)
                self.forsun.read_event.set()
                timer.stop()

        self.thread = threading.Thread(target=_)
        self.thread.setDaemon(True)
        self.thread.start()

    def stop(self):
        IOLoop.current().add_callback(lambda :IOLoop.current().stop())
        logging.info("server stoping")


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

import time
import subprocess
import logging
from tornado import gen
from ..action import Action, ExecuteActionError
from ... import config

class ShellAction(Action):
    def __init__(self, *args, **kwargs):
        super(ShellAction, self).__init__(*args, **kwargs)

        self.start_time = time.time()

    @gen.coroutine
    def execute(self, *args, **kwargs):
        if not self.params:
            raise ExecuteActionError("cmd is empty")

        cmd = self.params.get("cmd", 'ls .')
        cwd = self.params.get('cwd', config.get("ACTION_SHELL_CWD", '/tmp'))
        env = {}
        if self.params.get("env", ""):
            for e in self.params.get("env", "").split(";"):
                if e:
                    for k in e.split("="):
                        env[k[0]] = k[1]

        env["FORSUN_TIMESTAMP"] = str(self.ts)

        subprocess.Popen(cmd, shell=True, close_fds=True, cwd = cwd, env=env)
        logging.debug("shell action execute '%s' '%s' %.2fms", self.plan.key, cmd, (time.time() - self.start_time) * 1000)


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

from tornado import gen

class ExecuteActionError(Exception):
    pass

class Action(object):
    def __init__(self, ts, plan, action, params):
        self.ts = ts
        self.plan = plan
        self.action = action
        self.params = params

    @gen.coroutine
    def execute(self, *args, **kwargs):
        raise NotImplementedError()


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

from __future__ import absolute_import, division, print_function, with_statement

import os
import sys
import argparse
import multiprocessing
import atexit
from ..forsun import config
from ..utils import is_py3

parser = argparse.ArgumentParser(description='High-performance timing scheduling service')
parser.add_argument('--conf', dest='conf', default="", help='conf filename')
parser.add_argument('--bind', dest='bind_host', default="", help='bind host (default: 127.0.0.1)')
parser.add_argument('--port', dest='bind_port', default=0, type=int, help='bind port (default: 6458)')
parser.add_argument('--http', dest='http_bind', default="", help='bind http server (default: ) example: 127.0.0.1:80')
parser.add_argument('--demon', dest='demon', nargs='?', const=True, default=False, type=bool, help='run demon mode')
parser.add_argument('--nodemon', dest='nodemon', nargs='?', const=True, default=False, type=bool, help='run no demon mode')
parser.add_argument('--log', dest='log_file', default='', type=str, help='log file')
parser.add_argument('--log-level', dest='log_level', default='', type=str, help='log level (defaul: INFO)')
parser.add_argument('--driver', dest='driver', default='', type=str, help='store driver mem or redis (defaul: mem)')
parser.add_argument('--driver-mem-store-file', dest='store_mem_store_file', default='', type=str, help='store mem driver store file (defaul: ~/.forsun.dump)')
parser.add_argument('--driver-redis-host', dest='driver_redis_host', default='', type=str, help='store reids driver host (defaul: 127.0.0.1)')
parser.add_argument('--driver-redis-port', dest='driver_redis_port', default=0, type=int, help='store reids driver port (defaul: 6379)')
parser.add_argument('--driver-redis-db', dest='driver_redis_db', default=0, type=int, help='store reids driver db (defaul: 0)')
parser.add_argument('--driver-redis-password', dest='driver_redis_password', default='', type=str, help='store reids driver password (defaul: )')
parser.add_argument('--driver-redis-prefix', dest='driver_redis_prefix', default='', type=str, help='store reids driver key prefix (defaul: forsun)')
parser.add_argument('--driver-redis-server-id', dest='driver_redis_server_id', default=0, type=int, help='store reids driver server id (defaul: 0)')
parser.add_argument('--extension-path', dest='extension_path', default='', type=str, help='extension path')
parser.add_argument('--extension', dest='extensions', default=[], action="append", type=str, help='extension name')

def main():
    args = parser.parse_args()

    if args.conf:
        try:
            config.load_conf(args.conf)
        except Exception as e:
            print("load conf file error ", str(e))
            exit(1)

    if args.log_file:
        config.set("LOG_FILE", args.log_file)
    if args.log_level:
        config.set("LOG_LEVEL", args.log_level)

    if args.bind_host:
        config.set("BIND_ADDRESS", args.bind_host)
    if args.bind_port:
        config.set("PORT", args.bind_port)
    if args.http_bind:
        config.set("HTTP_BIND", args.http_bind)

    if args.driver:
        config.set("STORE_DRIVER", args.driver)
    if args.driver_redis_host:
        config.set("STORE_REDIS_HOST", args.driver_redis_host)
    if args.driver_redis_port:
        config.set("STORE_REDIS_PORT", args.driver_redis_port)
    if args.driver_redis_db:
        config.set("STORE_REDIS_DB", args.driver_redis_db)
    if args.driver_redis_password:
        config.set("STORE_REDIS_PASSWORD", args.driver_redis_password)
    if args.driver_redis_prefix:
        config.set("STORE_REDIS_PREFIX", args.driver_redis_prefix)
    if args.driver_redis_server_id:
        config.set("STORE_REDIS_SERVER_ID", args.driver_redis_server_id)

    if args.extension_path:
        config.set("EXTENSION_PATH", args.extension_path)
    if args.extensions:
        config.set("EXTENSIONS", args.extensions)


    if not args.nodemon:
        from ..forsun import Forsun

        def on_start(forsun):
            print("forsund started by pid %s" % p.pid)
            sys.stdin.close()
            sys.stdin = open(os.devnull)

            sys.stdout.close()
            sys.stdout = open(os.devnull, 'w')

            sys.stderr.close()
            sys.stderr = open(os.devnull, 'w')

        def run():
            try:
                forsun = Forsun()
                forsun.serve(on_start)
            except Exception as e:
                print(e)
                exit(1)

        p = multiprocessing.Process(target = run, name=" ".join(sys.argv))
        p.start()
        if is_py3:
            atexit._clear()
        else:
            atexit._exithandlers = []
    else:
        try:
            from ..forsun import Forsun
            forsun = Forsun()
            forsun.serve()
        except Exception as e:
            print(e)
            exit(1)

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
# 15/6/8
# create by: snower

import time
import signal
import traceback
import logging
import threading
from collections import deque
from .status import forsun_status

__time_out_callback = None
__exit_callback = None
__queues = deque()
__queue_ready_event = threading.Event()
__is_stop = False
__current_time = int(time.mktime(time.gmtime()))
__time_offset = int(time.time() - __current_time)

def exit_handler(signum, frame):
    __queues.append((__exit_callback, tuple()))
    if not __queue_ready_event.is_set():
        __queue_ready_event.set()
    else:
        forsun_status.timer_loop_ready_error_count += 1

def handler(signum, frame):
    global __current_time
    __current_time = int(time.time() - __time_offset)
    __queues.append((__time_out_callback, (__current_time,)))
    if not __queue_ready_event.is_set():
        __queue_ready_event.set()
    else:
        forsun_status.timer_loop_ready_error_count += 1

def reset():
    global __time_out_callback, __exit_callback, __queues, __queue_ready_event, __is_stop, __current_time
    __time_out_callback = None
    __exit_callback = None
    __queues = deque()
    __queue_ready_event = threading.Event()
    __is_stop = False
    __current_time = int(time.mktime(time.gmtime()))

def start(callback, exit_callback):
    global __time_out_callback, __exit_callback
    __time_out_callback = callback
    __exit_callback = exit_callback

def stop():
    global __time_out_callback, __exit_callback, __is_stop
    __time_out_callback = None
    __exit_callback = None
    __is_stop = True
    logging.info("timer stoping")

def loop():
    signal.signal(signal.SIGHUP, exit_handler)
    signal.signal(signal.SIGINT, exit_handler)
    signal.signal(signal.SIGTERM, exit_handler)
    signal.signal(signal.SIGALRM, handler)
    signal.setitimer(signal.ITIMER_REAL, 1, 1)
    logging.info("timer ready")
    while not __is_stop:
        try:
            if __queues:
                callback, args = __queues.popleft()
                if callback and callable(callback):
                    callback(*args)
                continue

            if __queue_ready_event.is_set():
                __queue_ready_event.clear()

            if not __queues:
                __queue_ready_event.wait(1.3)
            forsun_status.timer_loop_count += 1
        except KeyboardInterrupt:
            if __exit_callback and callable(__exit_callback):
                __exit_callback()
        except Exception as e:
            logging.info("timer error %s\n%s", e, traceback.format_exc())

            if __exit_callback and callable(__exit_callback):
                __exit_callback()

def current():
    return __current_time


# -*- coding: utf-8 -*-
# 15/6/8
# create by: snower

from .version import version, version_info

from . import config
from .plan import Plan
from .action.action import Action
from .store.store import Store
from .extension import Extension
from .clients import ThriftClient, TorThriftClient


# -*- coding: utf-8 -*-
# 18/1/26
# create by: snower

import logging
from .store import Store, register_store
from .action import Action, register_action

class ExtensionManager(type):
    extensions = []

    @classmethod
    def add_extension(cls, extension):
        cls.extensions.append(extension)

    @classmethod
    def get_extensions(cls):
        return cls.extensions

    @classmethod
    def register(cls):
        for extension_cls in cls.extensions:
            try:
                setattr(extension_cls, "_instance", extension_cls())
                extension_cls._instance.register()
            except Exception as e:
                logging.error("extension register error: %s %s %s", extension_cls, extension_cls.name, e)

    @classmethod
    def init(cls):
        for extension_cls in cls.extensions:
            if hasattr(extension_cls, "_instance") and extension_cls._instance:
                try:
                    extension_cls._instance.init()
                except Exception as e:
                    logging.error("extension init error: %s %s %s", extension_cls, extension_cls.name, e)

    @classmethod
    def uninit(cls):
        for extension_cls in cls.extensions:
            if hasattr(extension_cls, "_instance") and extension_cls._instance:
                try:
                    extension_cls._instance.uninit()
                    extension_cls._instance = None
                except Exception as e:
                    logging.error("extension uninit error: %s %s %s", extension_cls, extension_cls.name, e)

class Extension(object):
    name = ""

    def register_store(self, name, cls):
        return register_store(name, cls)

    def register_action(self, name, cls):
        return register_action(name, cls)

    def register(self):
        pass

    def init(self):
        pass

    def uninit(self):
        pass


# -*- coding: utf-8 -*-
# 18/1/23
# create by: snower

from .servers.processor.ttypes import ForsunPlanError

class ActionExecuteRetry(Exception):
    def __init__(self):
        pass

class UnknownError(ForsunPlanError):
    def __init__(self):
        super(UnknownError, self).__init__(1001, u"未知错误")

    def __hash__(self):
        return id(self)

class WillNeverArriveTimeError(ForsunPlanError):
    def __init__(self):
        super(WillNeverArriveTimeError, self).__init__(1002, u"定时时间错误")

    def __hash__(self):
        return id(self)

class NotFoundPlanError(ForsunPlanError):
    def __init__(self):
        super(NotFoundPlanError, self).__init__(1003, u"计划未找到")

    def __hash__(self):
        return id(self)

class StorePlanError(ForsunPlanError):
    def __init__(self):
        super(StorePlanError, self).__init__(1004, u"保存计划错误")

    def __hash__(self):
        return id(self)

class RemovePlanError(ForsunPlanError):
    def __init__(self):
        super(RemovePlanError, self).__init__(1005, u"移除计划错误")

    def __hash__(self):
        return id(self)

class UnknownActionError(ForsunPlanError):
    def __init__(self):
        super(UnknownActionError, self).__init__(1006, u"未知处理动作类型")

    def __hash__(self):
        return id(self)

class RequiredArgumentError(ForsunPlanError):
    def __init__(self, argument):
        super(RequiredArgumentError, self).__init__(1007, u"%s参数是必须的" % argument)

    def __hash__(self):
        return id(self)


# -*- coding: utf-8 -*-
# 15/6/8
# create by: snower

import logging
from tornado import ioloop
from tornado.ioloop import IOLoop, PeriodicCallback
from tornado import gen
from tornado.concurrent import Future
import tornadis
from ... import config
from ...plan import Plan
from ..store import Store
from ...utils import is_py3
from ... import timer

class HookPeriodicCallback(PeriodicCallback):
    def __init__(self, callback, callback_time, *args, **kwargs):
        super(HookPeriodicCallback, self).__init__(callback, callback_time)

ioloop.PeriodicCallback = HookPeriodicCallback

class RedisClient(object):
    def __init__(self, host, port, selected_db = 0, password= None, max_connections = 4, client_timeout = 7200, bulk_size = 5):
        self.ioloop = IOLoop.current()
        self.max_connections = max_connections
        self.current_connections = 0
        self.bulk_size = bulk_size
        self.pool = tornadis.ClientPool(
            max_size = -1,
            client_timeout = client_timeout,
            autoclose = True,
            host= host,
            port = port,
            db = selected_db,
            password = password,
            tcp_nodelay = True,
        )
        self._commands = []
        self.executing = False

    @gen.coroutine
    def execute(self):
        self.current_connections += 1
        try:
            connect_error_count = 0
            while self._commands:
                with (yield self.pool.connected_client()) as client:
                    if isinstance(client, Exception):
                        if self.current_connections != 1:
                            break

                        logging.error("redis store connect error: %s", client)
                        connect_error_count += 1
                        if connect_error_count <= 5:
                            yield gen.sleep(2)
                        else:
                            commands, self._commands = self._commands, []
                            for command in commands:
                                command[2].set_exception(client)
                        continue

                    if self._commands:
                        commands, self._commands = self._commands[:self.bulk_size], self._commands[self.bulk_size:]
                        if len(self._commands) > self.bulk_size and self.current_connections < self.max_connections:
                            self.ioloop.add_callback(self.execute)

                        try:
                            if len(commands) == 1:
                                reply = yield client.call(*commands[0][0], **commands[0][1])
                                if isinstance(reply, Exception):
                                    commands[0][2].set_exception(reply)
                                else:
                                    commands[0][2].set_result(reply)
                            else:
                                pipeline = tornadis.Pipeline()
                                for command in commands:
                                    pipeline.stack_call(*command[0])
                                replys = yield client.call(pipeline)
                                if isinstance(replys, Exception):
                                    for command in commands:
                                        command[2].set_exception(replys)
                                else:
                                    if isinstance(replys, (list, tuple)):
                                        for i in range(len(replys)):
                                            if isinstance(replys[i], Exception):
                                                commands[i][2].set_exception(replys[i])
                                            else:
                                                commands[i][2].set_result(replys[i])
                                    else:
                                        for command in commands:
                                            command[2].set_result(replys)
                        except Exception as e:
                            for command in commands:
                                if not command[2].done():
                                    command[2].set_exception(e)
        finally:
            self.current_connections -= 1
            if self.current_connections == 0:
                self.executing = False

    def execute_command(self, *args, **kwargs):
        future = Future()
        self._commands.append((args, kwargs, future))
        if not self.executing:
            self.ioloop.add_callback(self.execute)
            self.executing = True
        return future

    def select(self, db):
        return self.execute_command("SELECT", db)

    def get(self, key, **kwargs):
        return self.execute_command('GET', key, **kwargs)

    def set(self, key, value, expire=None, pexpire=None,
            only_if_not_exists=False, only_if_exists=False, **kwargs):
        args = []

        if expire is not None:
            args.extend(("EX", expire))
        if pexpire is not None:
            args.extend(("PX", pexpire))
        if only_if_not_exists:
            args.append("NX")
        if only_if_exists:
            args.append("XX")

        return self.execute_command('SET', key, value, *args, **kwargs)

    def keys(self, key, **kwargs):
        return self.execute_command('KEYS', key, **kwargs)

    def expire(self, key, ttl, **kwargs):
        return self.execute_command('EXPIRE', key, ttl, **kwargs)

    def delete(self, *keys, **kwargs):
        return self.execute_command('DEL', *keys, **kwargs)

    def hgetall(self, key, **kwargs):
        return self.execute_command('HGETALL', key, **kwargs)

    def hset(self, key, field, value, **kwargs):
        return self.execute_command('HSET', key, field, value, **kwargs)

    def hget(self, key, field, **kwargs):
        return self.execute_command('HGET', key, field, **kwargs)

    def hdel(self, key, *fields, **kwargs):
        return self.execute_command('HDEL', key, *fields, **kwargs)

class RedisStore(Store):
    def __init__(self, *args, **kwargs):
        super(RedisStore, self).__init__(*args, **kwargs)

        self.db = None
        self.prefix = "%s:%s" % (config.get("STORE_REDIS_PREFIX", "forsun"), config.get("STORE_REDIS_SERVER_ID", 0))

    @gen.coroutine
    def init(self):
        host = config.get("STORE_REDIS_HOST", "127.0.0.1")
        port = config.get("STORE_REDIS_PORT", 6379)
        selected_db = config.get("STORE_REDIS_DB", 0)
        password = config.get("STORE_REDIS_PASSWORD", None)

        self.db = RedisClient(
            host=host,
            port=port,
            selected_db=selected_db,
            password = password or None,
            max_connections=int(config.get("STORE_REDIS_MAX_CONNECTIONS", 8)),
            client_timeout=int(config.get("STORE_REDIS_CLIENT_TIMEOUT", 7200)),
            bulk_size=int(config.get("STORE_REDIS_BULK_SIZE", 5)),
        )
        logging.info("use redis store %s:%s/%s", host, port, selected_db)

    @gen.coroutine
    def set_current(self, current_time):
        res = yield self.db.set(self.prefix + ":current:time", str(current_time), expire = config.get("STORE_REDIS_CURRENTTIME_EXPRIED", 2592000))
        raise gen.Return(res)

    @gen.coroutine
    def get_current(self):
        res = yield self.db.get(self.prefix + ":current:time")
        raise gen.Return(int(res or 0))

    @gen.coroutine
    def set_plan(self, plan):
        res = yield self.db.set("".join([self.prefix, ":plan:", plan.key]), plan.dumps(), expire = plan.next_time - timer.current() + config.get("STORE_REDIS_PLAN_EXPRIED", 604800))
        raise gen.Return(res)

    @gen.coroutine
    def remove_plan(self, key):
        res = yield self.db.delete("".join([self.prefix, ":plan:", key]))
        raise gen.Return(res)

    @gen.coroutine
    def get_plan(self, key):
        res = yield self.db.get("".join([self.prefix, ":plan:", key]))
        if not res:
            raise gen.Return(None)
        try:
            res = Plan.loads(res)
        except Exception as e:
            logging.error("redis store load plan error: %s %s", key, e)
            res = None
        raise gen.Return(res)

    @gen.coroutine
    def add_time_plan(self, next_time, key):
        time_key = "".join([self.prefix, ":time:", str(next_time)])
        res = yield self.db.hset(time_key, key, '0')
        yield self.db.expire(time_key, next_time - timer.current() + config.get("STORE_REDIS_PLANTIME_EXPRIED", 604800))
        raise gen.Return(res)

    @gen.coroutine
    def get_time_plan(self, next_time, key):
        time_key = "".join([self.prefix, ":time:", str(next_time)])
        res = yield self.db.hget(time_key, key)
        try:
            status = int(res)
        except:
            status = 0
        raise gen.Return(status)

    @gen.coroutine
    def set_time_plan(self, next_time, key, status):
        time_key = "".join([self.prefix, ":time:", str(next_time)])
        res = yield self.db.hset(time_key, key, str(status))
        yield self.db.expire(time_key, next_time - timer.current() + config.get("STORE_REDIS_PLANTIME_EXPRIED", 604800))
        raise gen.Return(res)

    @gen.coroutine
    def remove_time_plan(self, next_time, key):
        res = yield self.db.hdel("".join([self.prefix, ":time:", str(next_time)]), key)
        raise gen.Return(res)

    @gen.coroutine
    def get_time_plans(self, ts):
        res = yield self.db.hgetall("".join([self.prefix, ":time:", str(ts)]))
        if not res:
            raise gen.Return([])
        if is_py3:
            raise gen.Return([str(res[i], "utf-8") for i in range(0, len(res), 2)])
        else:
            raise gen.Return([res[i] for i in range(0, len(res), 2)])

    @gen.coroutine
    def delete_time_plans(self, ts):
        res = yield self.db.delete("".join([self.prefix, ":time:", str(ts)]))
        raise gen.Return(res)

    @gen.coroutine
    def get_plan_keys(self, prefix = ""):
        prefix_len = len(self.prefix + ":plan:")
        res = yield self.db.keys("".join([self.prefix, ":plan:", prefix, "*"]))
        if is_py3:
            raise gen.Return([str(r, "utf-8")[prefix_len: ] for r in res])
        else:
            raise gen.Return([r[prefix_len: ] for r in res])


#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class ForsunPlanError(TException):
    """
    Attributes:
     - code
     - message
    """


    def __init__(self, code=None, message=None,):
        self.code = code
        self.message = message

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I16:
                    self.code = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.message = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ForsunPlanError')
        if self.code is not None:
            oprot.writeFieldBegin('code', TType.I16, 1)
            oprot.writeI16(self.code)
            oprot.writeFieldEnd()
        if self.message is not None:
            oprot.writeFieldBegin('message', TType.STRING, 2)
            oprot.writeString(self.message.encode('utf-8') if sys.version_info[0] == 2 else self.message)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ForsunPlan(object):
    """
    Attributes:
     - is_time_out
     - key
     - second
     - minute
     - hour
     - day
     - month
     - week
     - next_time
     - status
     - count
     - current_count
     - last_timeout
     - action
     - params
    """


    def __init__(self, is_time_out=None, key=None, second=None, minute=-1, hour=-1, day=-1, month=-1, week=-1, next_time=None, status=0, count=0, current_count=0, last_timeout=0, action="shell", params={
    },):
        self.is_time_out = is_time_out
        self.key = key
        self.second = second
        self.minute = minute
        self.hour = hour
        self.day = day
        self.month = month
        self.week = week
        self.next_time = next_time
        self.status = status
        self.count = count
        self.current_count = current_count
        self.last_timeout = last_timeout
        self.action = action
        if params is self.thrift_spec[15][4]:
            params = {
            }
        self.params = params

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.BOOL:
                    self.is_time_out = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.key = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I16:
                    self.second = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I16:
                    self.minute = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.I16:
                    self.hour = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.I16:
                    self.day = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.I16:
                    self.month = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.I16:
                    self.week = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 9:
                if ftype == TType.I32:
                    self.next_time = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 10:
                if ftype == TType.I16:
                    self.status = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 11:
                if ftype == TType.I16:
                    self.count = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 12:
                if ftype == TType.I16:
                    self.current_count = iprot.readI16()
                else:
                    iprot.skip(ftype)
            elif fid == 13:
                if ftype == TType.I32:
                    self.last_timeout = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 14:
                if ftype == TType.STRING:
                    self.action = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 15:
                if ftype == TType.MAP:
                    self.params = {}
                    (_ktype1, _vtype2, _size0) = iprot.readMapBegin()
                    for _i4 in range(_size0):
                        _key5 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        _val6 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.params[_key5] = _val6
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ForsunPlan')
        if self.is_time_out is not None:
            oprot.writeFieldBegin('is_time_out', TType.BOOL, 1)
            oprot.writeBool(self.is_time_out)
            oprot.writeFieldEnd()
        if self.key is not None:
            oprot.writeFieldBegin('key', TType.STRING, 2)
            oprot.writeString(self.key.encode('utf-8') if sys.version_info[0] == 2 else self.key)
            oprot.writeFieldEnd()
        if self.second is not None:
            oprot.writeFieldBegin('second', TType.I16, 3)
            oprot.writeI16(self.second)
            oprot.writeFieldEnd()
        if self.minute is not None:
            oprot.writeFieldBegin('minute', TType.I16, 4)
            oprot.writeI16(self.minute)
            oprot.writeFieldEnd()
        if self.hour is not None:
            oprot.writeFieldBegin('hour', TType.I16, 5)
            oprot.writeI16(self.hour)
            oprot.writeFieldEnd()
        if self.day is not None:
            oprot.writeFieldBegin('day', TType.I16, 6)
            oprot.writeI16(self.day)
            oprot.writeFieldEnd()
        if self.month is not None:
            oprot.writeFieldBegin('month', TType.I16, 7)
            oprot.writeI16(self.month)
            oprot.writeFieldEnd()
        if self.week is not None:
            oprot.writeFieldBegin('week', TType.I16, 8)
            oprot.writeI16(self.week)
            oprot.writeFieldEnd()
        if self.next_time is not None:
            oprot.writeFieldBegin('next_time', TType.I32, 9)
            oprot.writeI32(self.next_time)
            oprot.writeFieldEnd()
        if self.status is not None:
            oprot.writeFieldBegin('status', TType.I16, 10)
            oprot.writeI16(self.status)
            oprot.writeFieldEnd()
        if self.count is not None:
            oprot.writeFieldBegin('count', TType.I16, 11)
            oprot.writeI16(self.count)
            oprot.writeFieldEnd()
        if self.current_count is not None:
            oprot.writeFieldBegin('current_count', TType.I16, 12)
            oprot.writeI16(self.current_count)
            oprot.writeFieldEnd()
        if self.last_timeout is not None:
            oprot.writeFieldBegin('last_timeout', TType.I32, 13)
            oprot.writeI32(self.last_timeout)
            oprot.writeFieldEnd()
        if self.action is not None:
            oprot.writeFieldBegin('action', TType.STRING, 14)
            oprot.writeString(self.action.encode('utf-8') if sys.version_info[0] == 2 else self.action)
            oprot.writeFieldEnd()
        if self.params is not None:
            oprot.writeFieldBegin('params', TType.MAP, 15)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.params))
            for kiter7, viter8 in self.params.items():
                oprot.writeString(kiter7.encode('utf-8') if sys.version_info[0] == 2 else kiter7)
                oprot.writeString(viter8.encode('utf-8') if sys.version_info[0] == 2 else viter8)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.is_time_out is None:
            raise TProtocolException(message='Required field is_time_out is unset!')
        if self.key is None:
            raise TProtocolException(message='Required field key is unset!')
        if self.second is None:
            raise TProtocolException(message='Required field second is unset!')
        if self.next_time is None:
            raise TProtocolException(message='Required field next_time is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(ForsunPlanError)
ForsunPlanError.thrift_spec = (
    None,  # 0
    (1, TType.I16, 'code', None, None, ),  # 1
    (2, TType.STRING, 'message', 'UTF8', None, ),  # 2
)
all_structs.append(ForsunPlan)
ForsunPlan.thrift_spec = (
    None,  # 0
    (1, TType.BOOL, 'is_time_out', None, None, ),  # 1
    (2, TType.STRING, 'key', 'UTF8', None, ),  # 2
    (3, TType.I16, 'second', None, None, ),  # 3
    (4, TType.I16, 'minute', None, -1, ),  # 4
    (5, TType.I16, 'hour', None, -1, ),  # 5
    (6, TType.I16, 'day', None, -1, ),  # 6
    (7, TType.I16, 'month', None, -1, ),  # 7
    (8, TType.I16, 'week', None, -1, ),  # 8
    (9, TType.I32, 'next_time', None, None, ),  # 9
    (10, TType.I16, 'status', None, 0, ),  # 10
    (11, TType.I16, 'count', None, 0, ),  # 11
    (12, TType.I16, 'current_count', None, 0, ),  # 12
    (13, TType.I32, 'last_timeout', None, 0, ),  # 13
    (14, TType.STRING, 'action', 'UTF8', "shell", ),  # 14
    (15, TType.MAP, 'params', (TType.STRING, 'UTF8', TType.STRING, 'UTF8', False), {
    }, ),  # 15
)
fix_spec(all_structs)
del all_structs



# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

from tornado import gen

class Store(object):
    @gen.coroutine
    def init(self):
        pass

    @gen.coroutine
    def uninit(self):
        pass

    @gen.coroutine
    def set_current(self, current_time):
        raise NotImplemented

    @gen.coroutine
    def get_current(self):
        raise NotImplemented

    @gen.coroutine
    def set_plan(self, plan):
        raise NotImplementedError()

    @gen.coroutine
    def get_plan(self, plan):
        raise NotImplementedError()

    @gen.coroutine
    def remove_plan(self, key):
        raise NotImplementedError()

    @gen.coroutine
    def has_plan(self, key):
        plan = yield self.get_plan(key)
        raise gen.Return(plan)

    @gen.coroutine
    def add_time_plan(self, next_time, key):
        raise NotImplementedError()

    @gen.coroutine
    def get_time_plan(self, next_time, key):
        raise NotImplementedError()

    @gen.coroutine
    def set_time_plan(self, next_time, key, status):
        raise NotImplementedError()

    @gen.coroutine
    def remove_time_plan(self, next_time, key):
        raise NotImplementedError()

    @gen.coroutine
    def get_time_plans(self, ts):
        raise NotImplementedError()

    @gen.coroutine
    def delete_time_plans(self, ts):
        raise NotImplementedError()

    @gen.coroutine
    def get_plan_keys(self, prefix=""):
        raise NotImplementedError()


# -*- coding: utf-8 -*-
# 15/7/3
# create by: snower

import time
import logging
import socket
from tornado import gen
from tornado.iostream import IOStream
import beanstalkt
from ..action import Action
from ...utils import unicode_type
from ...error import ActionExecuteRetry

class BeanstalktClient(beanstalkt.Client):
    def _reconnect(self):
        pass

    @gen.coroutine
    def connect(self):
        """Connect to beanstalkd server."""
        if not self.closed():
            return
        self._talking = False
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
        self._stream = IOStream(self._socket)
        self._stream.set_close_callback(self._reconnect)
        yield self._stream.connect((self.host, self.port))

class BeanstalkAction(Action):
    client_pools = {}

    def __init__(self, *args, **kwargs):
        super(BeanstalkAction, self).__init__(*args, **kwargs)

        self.start_time  = time.time()

    @gen.coroutine
    def get_client(self, host, port, watch_key):
        key = "%s:%s:%s" % (host, port, watch_key)
        if key in self.client_pools:
            if self.client_pools[key].closed():
                del self.client_pools[key]

        if key not in self.client_pools:
            client = BeanstalktClient(host, port)
            yield client.connect()
            yield client.use(watch_key)
            if watch_key != "default":
                yield client.watch(watch_key)
                yield client.ignore("default")
            self.__class__.client_pools[key] = client
        raise gen.Return(self.client_pools[key])

    @gen.coroutine
    def execute(self, *args, **kwargs):
        host = self.params.get("host", "127.0.0.1")
        port = int(self.params.get("port", 11300))
        name = self.params.get('name', 'default')
        body = self.params.get("body", '')

        if not body:
            logging.error("beanstalk action execute error %s body is empty", self.plan.key)
            raise gen.Return(None)

        try:
            client = yield self.get_client(host, port, name)
            yield client.put(body.encode("utf-8") if isinstance(body, unicode_type) else body, ttr = 7200)
        except Exception as e:
            logging.error("beanstalk action execute error '%s' %s:%s '%s' '%s' '%s' %.2fms", self.plan.key, host,
                          port, name, body, e, (time.time() - self.start_time) * 1000)
            raise ActionExecuteRetry()
        else:
            logging.debug("beanstalk action execute '%s' %s:%s '%s' '%s' %.2fms", self.plan.key, host, port, name, body,
                          (time.time() - self.start_time) * 1000)


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

import time
import datetime
import struct
import msgpack
import pytz
from . import config

if config.get("STORE_DRIVER") == 'redis':
    server_id = int(config.get("STORE_REDIS_SERVER_ID"))
else:
    server_id = 0

class Plan(object):
    count_index = 0

    def __init__(self, key, second=0, minute = -1, hour = -1, day = -1, month = -1, week = -1, status = 0, count=0, is_time_out=False, next_time=None, current_count=0, last_timeout=0, created_time=0, action="event", params="{}"):
        self.key = key
        self.is_time_out = is_time_out
        self.second = second
        self.minute = minute
        self.hour = hour
        self.day = day
        self.month = month
        self.week = week
        self.status = status
        self.count = count
        self.current_count = int(current_count)
        self.last_timeout = int(last_timeout)
        self.created_time = float(created_time)
        self.action = action
        self.params = params

        if not self.key:
            self.key = self.gen_key()

        if self.is_time_out:
            self.get_timeout()
        else:
            self.get_plan()

        self.next_time = self.get_next_time() if not next_time or int(next_time) <= time.mktime(time.gmtime()) else int(next_time)

    def gen_key(self):
        self.__class__.count_index += 1
        return struct.pack("!III", self.created_time, server_id, self.count_index).encode("hex")

    def get_timeout(self):
        self.timeout_time = self.second if self.second >= 0 else 0
        self.timeout_time += self.minute * 60 if self.minute >= 0 else 0
        self.timeout_time += self.hour * 60 * 60 if self.hour >= 0 else 0
        self.timeout_time += self.day * 24 * 60 * 60 if self.day >= 0 else 0
        self.timeout_time += self.month * 30 * 24 * 60 * 60 if self.month >= 0 else 0
        self.timeout_time += self.week * 7 * 24 * 60 * 60 if self.week >= 0 else 0

    def get_plan(self):
        now = datetime.datetime.now(pytz.UTC)
        self.start_time = datetime.datetime(
            now.year,
            now.month if self.month == -1 else self.month,
            1 if self.day == -1 else self.day,
            0 if self.hour == -1 else self.hour,
            0 if self.minute == -1 else self.minute,
            0 if self.second == -1 else self.second,
            tzinfo=pytz.UTC,
        )

        self.step_time = None
        for pt in ["second", "minute", "hour", "day", "month"]:
            if getattr(self, pt) == -1:
                if pt != "month":
                    self.step_time = datetime.timedelta(**{pt+"s":1})
                else:
                    self.step_time = 1
                break

    def check(self, current_time):
        for pt in ["second", "minute", "hour", "day", "month"]:
            if getattr(self, pt) != -1 and getattr(self, pt) != getattr(current_time, pt):
                return False
        return True

    def get_next_time(self):
        if self.count != 0 and self.current_count >= self.count:
            return None

        if self.is_time_out:
            return int(time.mktime(time.gmtime())) + self.timeout_time
        else:
            now = datetime.datetime.now(pytz.UTC)
            if self.start_time >= now:
                return int(time.mktime(self.start_time.timetuple()))
            if self.start_time < now and self.step_time is None:
                return None
            current_time = self.start_time
            while not self.check(current_time) or current_time < now:
                if isinstance(self.step_time, int):
                    current_time = datetime.datetime(current_time.year if current_time.month < 12 else current_time.year + 1,
                                                     current_time.month + 1 if current_time.month < 12 else 1, current_time.day,
                                                     current_time.hour, current_time.minute, current_time.second,
                                                     tzinfo=pytz.UTC)
                else:
                    current_time += self.step_time
            return int(time.mktime(current_time.timetuple()))

    def dumps(self):
        return msgpack.dumps({
            "key": self.key,
            "second": self.second,
            "minute": self.minute,
            "hour": self.hour,
            "day": self.day,
            "month": self.month,
            "week": self.week,
            "status": self.status,
            "count": self.count,
            "is_time_out": self.is_time_out,
            "next_time": self.next_time,
            "current_count": self.current_count,
            "last_timeout": self.last_timeout,
            "created_time": self.created_time,
            "action": self.action,
            "params": self.params,
        })

    @classmethod
    def loads(cls, data):
        data = msgpack.loads(data, encoding = "utf-8")
        return cls(**data)

    def __str__(self):
        return "Plan('%s', %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)" % \
               (self.key, self.second, self.minute, self.hour, self.day, self.month, self.week, self.status,
                self.count, self.is_time_out, self.next_time, self.current_count, self.last_timeout)


# -*- coding: utf-8 -*-
# 15/7/3
# create by: snower

import time
import logging
from tornado.ioloop import IOLoop
from tornado import gen
from tornado.concurrent import Future
import tornadis
from ..action import Action, ExecuteActionError
from ...utils import parse_cmd
from ... import config
from ...error import ActionExecuteRetry

class RedisClient(object):
    def __init__(self, host, port, selected_db = 0, password = None, max_connections = 4, client_timeout = 7200, bulk_size = 5):
        self.ioloop = IOLoop.current()
        self.max_connections = max_connections
        self.current_connections = 0
        self.bulk_size = bulk_size
        self.pool = tornadis.ClientPool(
            max_size = -1,
            client_timeout = client_timeout,
            autoclose = True,
            host= host,
            port = port,
            db = selected_db,
            password = password,
            tcp_nodelay = True,
        )
        self._commands = []
        self.executing = False

    @gen.coroutine
    def execute(self):
        self.current_connections += 1
        try:
            connect_error_count = 0
            while self._commands:
                with (yield self.pool.connected_client()) as client:
                    if isinstance(client, Exception):
                        if self.current_connections != 1:
                            break

                        logging.error("redis store connect error: %s", client)
                        connect_error_count += 1
                        if connect_error_count <= 2:
                            yield gen.sleep(2)
                        else:
                            commands, self._commands = self._commands, []
                            for command in commands:
                                command[2].set_exception(client)
                        continue

                    if self._commands:
                        commands, self._commands = self._commands[:self.bulk_size], self._commands[self.bulk_size:]
                        if len(self._commands) > self.bulk_size and self.current_connections < self.max_connections:
                            self.ioloop.add_callback(self.execute)

                        try:
                            if len(commands) == 1:
                                reply = yield client.call(*commands[0][0], **commands[0][1])
                                if isinstance(reply, Exception):
                                    commands[0][2].set_exception(reply)
                                else:
                                    commands[0][2].set_result(reply)
                            else:
                                pipeline = tornadis.Pipeline()
                                for command in commands:
                                    pipeline.stack_call(*command[0])
                                replys = yield client.call(pipeline)
                                if isinstance(replys, Exception):
                                    for command in commands:
                                        command[2].set_exception(replys)
                                else:
                                    if isinstance(replys, (list, tuple)):
                                        for i in range(len(replys)):
                                            if isinstance(replys[i], Exception):
                                                commands[i][2].set_exception(replys[i])
                                            else:
                                                commands[i][2].set_result(replys[i])
                                    else:
                                        for command in commands:
                                            command[2].set_result(replys)
                        except Exception as e:
                            for command in commands:
                                if not command[2].done():
                                    command[2].set_exception(e)
        finally:
            self.current_connections -= 1
            if self.current_connections == 0:
                self.executing = False

    def execute_command(self, *args, **kwargs):
        future = Future()
        self._commands.append((args, kwargs, future))
        if not self.executing:
            self.ioloop.add_callback(self.execute)
            self.executing = True
        return future

class RedisAction(Action):
    client_pools = {}

    def __init__(self, *args, **kwargs):
        super(RedisAction, self).__init__(*args, **kwargs)

        self.start_time  = time.time()

    def get_client(self, host, port, selected_db, password, max_connections):
        key = "%s:%s:%s" % (host, port, selected_db)
        if key not in self.client_pools:
            client_timeout = int(config.get("ACTION_REDIS_CLIENT_TIMEOUT", 8))
            bulk_size = int(config.get("ACTION_REDIS_BULK_SIZE", 8))
            self.__class__.client_pools[key] = RedisClient(host = host, port = port, selected_db = selected_db, max_connections = max_connections,
                                                           client_timeout = client_timeout, bulk_size = bulk_size)
        return self.__class__.client_pools[key]

    @gen.coroutine
    def execute(self, *args, **kwargs):
        if len(self.params) < 2:
            raise ExecuteActionError("redis params is empty")

        host = self.params.get("host", "127.0.0.1")
        port = int(self.params.get("port", 6379))
        selected_db = int(self.params.get("selected_db", 0))
        password = self.params.get("password", None)
        max_connections = int(self.params.get("max_connections", config.get("ACTION_REDIS_MAX_CONNECTIONS", 8)))
        command = self.params.get("command")
        cmds = parse_cmd(command, True)

        if cmds:
            try:
                client = self.get_client(host, port, selected_db, password, max_connections)
                try:
                    futures = []
                    for cmd, args in cmds:
                        future = client.execute_command(cmd, *args)
                        futures.append(future)
                    yield futures
                except Exception as e:
                    logging.debug("redis action execute '%s' %s:%s/%s '%s' %s %.2fms", self.plan.key, host, port,
                                  selected_db, cmds, e, (time.time() - self.start_time) * 1000)
            except Exception as e:
                logging.debug("redis action execute '%s' %s:%s/%s '%s' %s %.2fms", self.plan.key, host, port,
                              selected_db, cmds, e, (time.time() - self.start_time) * 1000)
                raise ActionExecuteRetry()
        logging.debug("redis action execute '%s' %s:%s/%s '%s' %.2fms", self.plan.key, host, port, selected_db,
                      cmds, (time.time() - self.start_time) * 1000)


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

import time
import logging
from tornado import gen
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from ..action import Action, ExecuteActionError
from ... import config
from ...error import ActionExecuteRetry

class HttpAction(Action):
    def __init__(self, *args, **kwargs):
        super(HttpAction, self).__init__(*args, **kwargs)

        self.client = AsyncHTTPClient()
        self.start_time = time.time()
        self.configed = False

    def config(self):
        import forsun
        AsyncHTTPClient.configure(None,
                                  max_clients=config.get("ACTION_HTTP_MAX_CLIENTS", 64),
                                  defaults={"user_agent": config.get("ACTION_HTTP_USER_AGENT") or ("forsun/%s" % forsun.version)
                                  })
        self.configed = True

    @gen.coroutine
    def execute(self, *args, **kwargs):
        if not self.configed:
            self.config()

        if not self.params:
            raise ExecuteActionError("url is empty")

        url = self.params["url"]
        method = self.params.get("method", 'GET')
        if method.lower() not in ("get", "post", "put", "delete", "head"):
            method= 'GET'

        body = self.params.get('body', None)
        headers = {}
        for key in self.params:
            if key.startswith("header_"):
                headers[key[7:]] = self.params[key]

        auth_username = self.params.get('auth_username', None)
        auth_password = self.params.get('auth_password', None)
        auth_mode = self.params.get('auth_mode', None)
        user_agent = self.params.get("user_agent", None)
        connect_timeout = self.params.get("connect_timeout", config.get("ACTION_HTTP_CONNECT_TIMEOUT", 5))
        request_timeout = self.params.get("request_timeout", config.get("ACTION_HTTP_REQUEST_TIMEOUT", 120))

        headers["X-FORSUN-TIMESTAMP"] = str(self.ts)
        request = HTTPRequest(url, method, body=body, headers=headers,
                              auth_username=auth_username, auth_password=auth_password, auth_mode=auth_mode,
                              user_agent = user_agent, connect_timeout=connect_timeout, request_timeout=request_timeout
        )
        response = yield self.client.fetch(request, raise_error = False)
        logging.debug("http action execute '%s' %s '%s' %s '%s' %.2fms", self.plan.key, method, url,
                      response.code, response.reason, (time.time() - self.start_time) * 1000)
        if response and response.code >= 500:
            raise ActionExecuteRetry()


# -*- coding: utf-8 -*-
# 15/6/10
# create by: snower

import os
import sys
from setuptools import find_packages, setup

version = "0.1.3"

if os.path.exists("README.rst"):
    if sys.version_info[0] >= 3:
        with open("README.rst", encoding="utf-8") as fp:
            long_description = fp.read()
    else:
        with open("README.rst") as fp:
            long_description = fp.read()
else:
    long_description = ''

setup(
    name='forsun',
    version=version,
    url='https://github.com/snower/forsun',
    author='snower',
    author_email='sujian199@gmail.com',
    license='MIT',
    packages=find_packages(),
    zip_safe=False,
    install_requires=[
        'tornado>=5.0, <=5.1',
        'thrift==0.11.0',
        'torthrift>=0.2.4',
        'tornadis>=0.8.0, <=0.8.1',
        'greenlet>=0.4.2',
        'msgpack>=0.5.1',
        'pytz>=2017.3',
        'tzlocal>=1.5.1',
        'configparser>=3.5.0',
        'psutil>=5.4.7',
    ],
    extras_require = {
        'mysql': ['tormysql>=0.3.7'],
        'beanstalk': ['beanstalkt>=0.7.0'],
    },
    package_data={
        '': ['README.md'],
    },
    entry_points={
        'console_scripts': [
            'forsun = forsun.scripts.forsun:main',
            'forsund = forsun.scripts.forsund:main',
        ],
    },
    description= 'High-performance high-precision timing scheduling service',
    long_description= long_description,
)



# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

import time
import logging
from tornado import gen
from ..plan import Plan
from ..status import forsun_status
from .processor.ttypes import ForsunPlan, ForsunPlanError

class Handler(object):
    def __init__(self, forsun):
        self.forsun = forsun

    def to_eplan(self, plan):
        plan = ForsunPlan(
            is_time_out=plan.is_time_out,
            key=plan.key,
            second=plan.second,
            minute=plan.minute,
            hour=plan.hour,
            day=plan.day,
            month=plan.month,
            week=plan.week,
            next_time=plan.next_time,
            status=plan.status,
            count=plan.count,
            current_count=plan.current_count,
            action=plan.action,
            params=plan.params,
        )
        return plan

    @gen.coroutine
    def ping(self):
        raise gen.Return(0)

    @gen.coroutine
    def create(self, key, second, minute = -1, hour = -1, day = -1, month = -1, week = -1, action="shell", params=None):
        start_time = time.time()
        forsun_status.requesting_count += 1
        try:
            plan = Plan(key, second, minute, hour, day, month, week, is_time_out=False, action=action,
                        params=params or {}, created_time=time.mktime(time.gmtime()))
            yield self.forsun.create_plan(plan)
            forsun_status.created_count += 1
        except Exception as e:
            forsun_status.requested_error_count += 1
            logging.debug("create error %s %s %s %s %s %s %s %s %.2fms %s", key, second, minute, hour, day, month, week, action,
                          (time.time() - start_time) * 1000, e)
            raise e
        else:
            logging.debug("create %s %s %s %s %s %s %s %s %.2fms", key, second, minute, hour, day, month, week, action,
                          (time.time() - start_time) * 1000)
        finally:
            forsun_status.requested_count += 1
            forsun_status.requesting_count -= 1
        raise gen.Return(self.to_eplan(plan))

    @gen.coroutine
    def createTimeout(self, key, second, minute = -1, hour = -1, day = -1, month = -1, week = -1, count=0, action="shell", params=None):
        start_time = time.time()
        forsun_status.requesting_count += 1
        try:
            plan = Plan(key, second, minute, hour, day, month, week, is_time_out=True, count=count, action=action,
                        params=params or {}, created_time=time.mktime(time.gmtime()))
            yield self.forsun.create_plan(plan)
            forsun_status.create_timeouted_count += 1
        except Exception as e:
            forsun_status.requested_error_count += 1
            logging.debug("createTimeout error %s %s %s %s %s %s %s %s %s %.2fms %s", key, second, minute, hour, day, month,
                          week, count, action, (time.time() - start_time) * 1000, e)
            raise e
        else:
            logging.debug("createTimeout %s %s %s %s %s %s %s %s %s %.2fms", key, second, minute, hour, day, month,
                          week, count, action, (time.time() - start_time) * 1000)
        finally:
            forsun_status.requested_count += 1
            forsun_status.requesting_count -= 1
        raise gen.Return(self.to_eplan(plan))

    @gen.coroutine
    def remove(self, key):
        start_time = time.time()
        forsun_status.requesting_count += 1
        try:
            plan = yield self.forsun.remove_plan(key)
            forsun_status.removed_count += 1
        except Exception as e:
            forsun_status.requested_error_count += 1
            logging.debug("remove error %s %.2fms %s", key, (time.time() - start_time) * 1000, e)
            raise e
        else:
            logging.debug("remove %s %.2fms", key, (time.time() - start_time) * 1000)
        finally:
            forsun_status.requested_count += 1
            forsun_status.requesting_count -= 1
        raise gen.Return(self.to_eplan(plan))

    @gen.coroutine
    def get(self, key):
        start_time = time.time()
        forsun_status.requesting_count += 1
        try:
            plan = yield self.forsun.get_pan(key)
        except Exception as e:
            forsun_status.requested_error_count += 1
            logging.debug("get error %s %.2fms %s", key, (time.time() - start_time) * 1000, e)
            raise e
        else:
            logging.debug("get %s %.2fms", key, (time.time() - start_time) * 1000)
        finally:
            forsun_status.requested_count += 1
            forsun_status.requesting_count -= 1
        raise gen.Return(self.to_eplan(plan))

    @gen.coroutine
    def getCurrent(self):
        start_time = time.time()
        forsun_status.requesting_count += 1
        try:
            current_ts = yield self.forsun.get_current_time()
            plans = yield self.forsun.get_time_plans(current_ts)
        except Exception as e:
            forsun_status.requested_error_count += 1
            logging.debug("getCurrent error %.2fms %s", (time.time() - start_time) * 1000, e)
            raise e
        else:
            logging.debug("getCurrent %.2fms", (time.time() - start_time) * 1000)
        finally:
            forsun_status.requested_count += 1
            forsun_status.requesting_count -= 1
        plans = [self.to_eplan(plan) for plan in plans]
        raise gen.Return(plans)

    @gen.coroutine
    def getTime(self, timestamp):
        start_time = time.time()
        forsun_status.requesting_count += 1
        try:
            plans = yield self.forsun.get_time_plans(timestamp)
        except Exception as e:
            forsun_status.requested_error_count += 1
            logging.debug("getTime error %.2fms %s", (time.time() - start_time) * 1000, e)
            raise e
        else:
            logging.debug("getTime %.2fms", (time.time() - start_time) * 1000)
        finally:
            forsun_status.requested_count += 1
            forsun_status.requesting_count -= 1
        plans = [self.to_eplan(plan) for plan in plans]
        raise gen.Return(plans)

    @gen.coroutine
    def getKeys(self, prefix=""):
        start_time = time.time()
        forsun_status.requesting_count += 1
        try:
            keys = yield self.forsun.get_keys(prefix)
        except Exception as e:
            forsun_status.requested_error_count += 1
            logging.debug("getKeys error %.2fms", (time.time() - start_time) * 1000, e)
            raise e
        else:
            logging.debug("getKeys %.2fms", (time.time() - start_time) * 1000)
        finally:
            forsun_status.requested_count += 1
            forsun_status.requesting_count -= 1
        raise gen.Return(keys)

    @gen.coroutine
    def info(self):
        start_time = time.time()
        forsun_status.requesting_count += 1
        try:
            info = forsun_status.get_info()
        except Exception as e:
            forsun_status.requested_error_count += 1
            logging.debug("info error %.2fms", (time.time() - start_time) * 1000, e)
            raise e
        else:
            logging.debug("info %.2fms", (time.time() - start_time) * 1000)
        finally:
            forsun_status.requested_count += 1
            forsun_status.requesting_count -= 1
        raise gen.Return(info)


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

import os
import configparser
from .utils import unicode_type, string_type, number_type, ensure_unicode

class ConfFileNotFoundError(Exception):
    pass

__config = {}

DEFAULT_CONFIG = {
    "LOG_FILE": "/var/log/forsun.log",
    "LOG_LEVEL": "INFO",
    "LOG_FORMAT": "%(asctime)s %(process)d %(levelname)s %(message)s",
    "LOG_ROTATE": "",
    "LOG_BACKUP_COUNT": 64,

    "BIND_ADDRESS": "127.0.0.1",
    "PORT": 6458,

    "HTTP_BIND": "",
    "SERVER_MODE": "TIMER",

    "STORE_DRIVER": "mem",
    "STORE_STATUS_EXPRIED": 0,

    "STORE_MEM_STORE_FILE": "",
    "STORE_MEM_TIME_RATE": 1,

    "STORE_REDIS_HOST": "127.0.0.1",
    "STORE_REDIS_PORT": 6379,
    "STORE_REDIS_DB": 0,
    "STORE_REDIS_PASSWORD": "",
    "STORE_REDIS_PREFIX": "forsun",
    "STORE_REDIS_SERVER_ID": 0,
    "STORE_REDIS_MAX_CONNECTIONS": 8,
    "STORE_REDIS_CLIENT_TIMEOUT": 7200,
    "STORE_REDIS_BULK_SIZE": 5,
    "STORE_REDIS_CURRENTTIME_EXPRIED": 2592000,
    "STORE_REDIS_PLAN_EXPRIED": 604800,
    "STORE_REDIS_PLANTIME_EXPRIED": 604800,

    "ACTION_RETRY_DELAY_SECONDS": 3,
    "ACTION_RETRY_DELAY_RATE": 1,
    "ACTION_RETRY_MAX_COUNT": 3,
    "ACTION_POLL_IDLE_SECONDS": 120,
    "ACTION_SHELL_CWD": "/tmp",
    "ACTION_HTTP_MAX_CLIENTS": 64,
    "ACTION_HTTP_CONNECT_TIMEOUT": 5,
    "ACTION_HTTP_REQUEST_TIMEOUT": 120,
    "ACTION_HTTP_USER_AGENT": "",
    "ACTION_REDIS_MAX_CONNECTIONS": 8,
    "ACTION_REDIS_CLIENT_TIMEOUT": 7200,
    "ACTION_REDIS_BULK_SIZE": 5,
    "ACTION_THRIFT_MAX_CONNECTIONS": 64,
    "ACTION_MYSQL_USER": "root",
    "ACTION_MYSQL_PASSWD": "",
    "ACTION_MYSQL_MAX_CONNECTIONS": 8,
    "ACTION_MYSQL_WAIT_CONNECTION_TIMEOUT": 7200,
    "ACTION_MYSQL_IDLE_SECONDS": 120,

    "EXTENSION_PATH": "",
    "EXTENSIONS": [],
}

def get(name, default=None):
    return __config.get(name, default)

def set(name, value):
    old_value = __config[name]
    __config[name] = value
    return old_value

def update(config):
    __config.update(config)
    return __config

update(DEFAULT_CONFIG)
for key, value in DEFAULT_CONFIG.items():
    env_value = os.environ.get(key)
    if env_value is not None:
        try:
            if isinstance(value, number_type):
                set(key, int(env_value))
            elif isinstance(value, float):
                set(key, float(env_value))
            elif isinstance(value, string_type):
                set(key, str(env_value))
        except:pass

def load_conf(filename):
    try:
        with open(filename, "r") as fp:
            conf_content = unicode_type("[global]\n") + ensure_unicode(fp.read())
            cf = configparser.ConfigParser(allow_no_value=True)
            cf.read_string(conf_content)

            for key, value in DEFAULT_CONFIG.items():
                try:
                    if key.startswith("STORE_"):
                        conf_value = cf.get("store", key[6:].lower())
                    elif key.startswith("ACTION_"):
                        conf_value = cf.get("action", key[7:].lower())
                    elif key.startswith("EXTENSION"):
                        if key == "EXTENSIONS":
                            conf_value = cf.get("extension", "extensions")
                            if isinstance(conf_value, string_type):
                                set(key, conf_value.split(";"))
                                continue
                        else:
                            conf_value = cf.get("extension", key[10:].lower())
                    else:
                        conf_value = cf.get("global", key.lower())

                    try:
                        if isinstance(value, number_type):
                            set(key, int(conf_value))
                        elif isinstance(value, float):
                            set(key, float(conf_value))
                        elif isinstance(value, string_type):
                            set(key, str(conf_value))
                    except:
                        pass
                except (configparser.NoOptionError, configparser.NoSectionError):
                    continue
    except IOError:
        raise ConfFileNotFoundError()



# -*- coding: utf-8 -*-
# 15/7/3
# create by: snower

import time
import logging
from tornado import gen
from thrift.protocol.TBinaryProtocol import TBinaryProtocolFactory
from thrift.Thrift import TApplicationException
from torthrift.pool import TStreamPool
from torthrift.client import PoolClient
from ...clients.client.Forsun import Client
from ..action import Action, ExecuteActionError
from ... import config
from ...error import ActionExecuteRetry

class ThriftAction(Action):
    client_pools = {}

    def __init__(self, *args, **kwargs):
        super(ThriftAction, self).__init__(*args, **kwargs)

        self.start_time  = time.time()

    def get_client(self, host, port, max_connections = 0):
        key = "%s:%s" % (host, port)
        if key not in self.client_pools:
            self.__class__.client_pools[key] = PoolClient(Client, TStreamPool(host, port, max_stream=max_connections), TBinaryProtocolFactory())
        elif max_connections:
            self.client_pools[key]._itrans_pool._max_stream = max_connections
        return self.client_pools[key]

    @gen.coroutine
    def execute(self, *args, **kwargs):
        if not self.params:
            raise ExecuteActionError("is empty")

        host = self.params.get("host", "127.0.0.1")
        port = int(self.params.get("port", 5643))
        max_connections = int(self.params.get("max_connections", config.get("ACTION_THRIFT_MAX_CONNECTIONS", 64)))

        try:
            client = self.get_client(host, port, max_connections)
            try:
                yield client.forsun_call(self.plan.key, int(self.ts), self.params)
            except TApplicationException as e:
                logging.error("thrift action execute error '%s' %s:%s '%s' %.2fms", self.plan.key, host, port, e,
                              (time.time() - self.start_time) * 1000)
            else:
                logging.debug("thrift action execute '%s' %s:%s %.2fms", self.plan.key, host, port,
                              (time.time() - self.start_time) * 1000)
        except Exception as e:
            logging.error("thrift action execute error '%s' %s:%s '%s' %.2fms", self.plan.key, host, port, e,
                          (time.time() - self.start_time) * 1000)
            raise ActionExecuteRetry()